from __future__ import annotations

from contextlib import nullcontext
from pathlib import Path
from typing import Dict, List, Tuple, Union

import rich.console
import rich.layout
import rich.live
import rich.panel
import rich.tree
from PIL import Image
from typing_extensions import Self

from verdict.core.executor import GraphExecutor
from verdict.core.primitive import Block, Layer, MaterializationContext, Unit
from verdict.core.visualization import StreamingLayoutManager
from verdict.dataset import DatasetWrapper
from verdict.model import ModelSelectionPolicy
from verdict.schema import Schema
from verdict.util.exceptions import VerdictDeclarationTimeError
from verdict.util.log import init_logger, logger
from verdict.util.misc import keyboard_interrupt_safe


class Pipeline:
    name: str
    block: Block

    executor: GraphExecutor

    def __init__(self, name: str='Pipeline') -> None:
        super().__init__()
        self.name = name
        self.block = Block()

    def copy(self) -> "Pipeline":
        pipeline = Pipeline(self.name)
        pipeline.block = self.block.copy()
        return pipeline

    def __rshift__(self, other: Union[Unit, Layer, "Block"]) -> Self:
        self.block >>= other
        return self

    def via(self, policy_or_name: Union[ModelSelectionPolicy, str], retries: int=1, **inference_parameters) -> Self:
        self.block.via(policy_or_name, retries, **inference_parameters)
        return self

    def collect_outputs(self, executor: GraphExecutor, block_instance: Block) -> Tuple[Dict[str, Schema], List[str]]:
        leaf_node_prefixes: List[str] = []
        outputs: Dict[str, Schema] = {}

        for node in block_instance.nodes:
            try:
                if node not in executor.outputs:
                    logger.debug(f"Node {node} has no output!")
                    continue

                if not isinstance(executor.outputs[node], Schema): # type: ignore
                    logger.debug(f"Node {node} with output {executor.outputs[node]} is not a Schema") # type: ignore

                for key, value in executor.outputs[node].model_dump().items(): # type: ignore
                    outputs[(column_name := f"{self.name}_{'.'.join(node.prefix)}_{key}")] = value # type: ignore
                    if node in block_instance.leaf_nodes:
                        leaf_node_prefixes.append(column_name)
            except Exception as e:
                raise VerdictDeclarationTimeError(f"Error collecting outputs for node {node} with outputs {executor.outputs[node]}") from e # type: ignore

        return outputs, sorted(list(leaf_node_prefixes))

    @keyboard_interrupt_safe
    def run(
        self,
        input_data: Schema=Schema.empty(),
        max_workers: int=128,
        display: bool=False,
        graceful: bool=False
    ) -> Tuple[Dict[str, Schema], List[str]]:
        self.block = self.block.copy()

        init_logger(self.name)
        logger.info(f"Starting pipeline {self.name}")
        self.executor = GraphExecutor(max_workers=max_workers)

        with (rich.live.Live(auto_refresh=True) if display else nullcontext()) as live: # type: ignore
            context = MaterializationContext(n_instances=1)
            if display:
                context.parent_branch = rich.tree.Tree(self.name)
                context.streaming_layout_manager = StreamingLayoutManager(rich.layout.Layout())

                layout = rich.layout.Layout()
                layout.split_row(
                    rich.layout.Layout(rich.panel.Panel(context.parent_branch, title="Execution Tree"), name="execution"),
                    rich.layout.Layout(rich.panel.Panel(context.streaming_layout_manager.layout, title="Streaming"), name="streaming")
                )

                live.update(layout) # type: ignore

            block_instance, _ = self.block._materialize(context.copy())
            block_instance.source_input = input_data
            block_instance.executor = self.executor

            self.executor.submit(block_instance.root_nodes, input_data) # type: ignore
            self.executor.wait_for_completion(graceful=graceful)

        logger.info(f"Pipeline {self.name} completed")
        return self.collect_outputs(self.executor, block_instance)

    @keyboard_interrupt_safe
    def run_from_dataset(self,
            dataset: DatasetWrapper,
            max_workers: int=128,
            experiment_config=None, # verdict.util.experiment.ExperimentConfig
            display: bool=False,
            graceful: bool=False
        ) -> Tuple["pd.DataFrame", List[str]]: # type: ignore
        self.block = self.block.copy()

        init_logger(self.name)
        logger.info(f"Running pipeline {self.name} on dataset (len={len(dataset)})")
        self.executor = GraphExecutor(max_workers=max_workers)

        dataset_df = dataset.samples.copy()

        with (rich.live.Live(auto_refresh=True) if display else nullcontext()) as live: # type: ignore
            block_instances: Dict[str, Block] = {}

            context = MaterializationContext(n_instances=len(dataset))
            if display:
                context.parent_branch = rich.tree.Tree(self.name)
                context.streaming_layout_manager = StreamingLayoutManager(rich.layout.Layout())

                execution_layout = rich.layout.Layout(name="execution")
                execution_tree_layout = rich.layout.Layout(rich.panel.Panel(context.parent_branch, title="Execution Tree"), ratio=1, name="execution")
                if experiment_config:
                    from verdict.util.experiment import get_experiment_layout
                    execution_layout.split_column(
                        get_experiment_layout(dataset_df, experiment_config),
                        execution_tree_layout
                    )
                else:
                    execution_layout.add_split(execution_tree_layout)

                layout = rich.layout.Layout()
                layout.split_row(
                    execution_layout,
                    rich.layout.Layout(rich.panel.Panel(context.streaming_layout_manager.layout, title="Streaming"), name="streaming")
                )

                live.update(layout) # type: ignore

            prototype, _ = self.block._materialize(context)

            for idx, (row, input_data) in enumerate(dataset):
                block_instance = block_instances[row['hash(row)']] = prototype.clone()
                block_instance.source_input = input_data
                block_instance.executor = self.executor

                self.executor.submit(block_instance.root_nodes, input_data, leader=(idx == 0)) # type: ignore

            self.executor.wait_for_completion(graceful=graceful)

            output = []
            leaf_node_prefixes = set()
            for row_id, block_instance in block_instances.items():
                row_output = {"hash(row)": row_id}
                outputs, _leaf_node_prefixes = self.collect_outputs(self.executor, block_instance)

                row_output.update(outputs)
                leaf_node_prefixes.update(_leaf_node_prefixes)

                if len(row_output) > 1:
                    output.append(row_output)

            if len(output) > 0:
                import pandas as pd
                result_df = pd.merge(
                    dataset_df,
                    pd.DataFrame(output),
                    on="hash(row)",
                    how="right"
                ).drop(columns=["hash(row)"])
            else:
                result_df = dataset_df

            if display and experiment_config:
                from verdict.util.experiment import get_experiment_layout
                execution_layout["experiment"].update(get_experiment_layout(result_df, experiment_config))
                live.refresh() # type: ignore

            return result_df, sorted(list(leaf_node_prefixes))

    @keyboard_interrupt_safe
    def run_from_list(
        self,
        dataset: List[Schema],
        max_workers: int = 128,
        experiment_config=None, # verdict.util.experiment.ExperimentConfig
        display: bool = False,
        graceful: bool = False,
    ) -> Tuple[Dict[str, Schema], List[str]]:
        from datasets import Dataset  # type: ignore[import-untyped]
        vedict_dataset = DatasetWrapper(
            Dataset.from_list([data.model_dump() for data in dataset])
        )
        return self.run_from_dataset(
            vedict_dataset, max_workers, experiment_config, display, graceful
        )

    def checkpoint(self, path: Path):
        # TODO
        pass

    def restore(self, path: Path):
        # TODO
        pass

    def plot(self, display=False) -> Image.Image:
        return self.block.materialize().plot(display)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import dill  # type: ignore[import-untyped]
from typing_extensions import Self

from verdict.schema import Schema
from verdict.util.exceptions import ConfigurationError

InputFn = Callable[[Dict[str, Any]], Schema]

class DatasetWrapper(Iterator[Tuple[Dict[str, Any], Schema]]):
    dataset: "pd.DataFrame" # type: ignore
    samples: "pd.DataFrame" # type: ignore

    max_samples: Optional[int]

    input_fn: InputFn

    _iter: Optional[Iterator[Tuple[Dict[str, Any], Schema]]] = None
    _count: int = 0

    def __init__(
            self,
            dataset: "Dataset", # type: ignore
            input_fn: Optional[InputFn]=None,
            columns: Optional[List[str]]=None,
            max_samples: Optional[int]=None,
    ):
        import pandas as pd
        self.dataset = pd.DataFrame(dataset)
        self.dataset["hash(row)"] = self.dataset.apply(lambda row: hash(str(row)), axis=1)
        self.max_samples = int(max_samples) if max_samples else None
        self.samples = self.dataset.sample(n=self.max_samples) if self.max_samples else self.dataset

        if input_fn is not None and columns is not None:
            raise ConfigurationError("Cannot specify both input_fn and columns")

        columns = columns or (self.dataset.columns.difference(["hash(row)"]).tolist())
        self.input_fn = input_fn or (lambda row: Schema.of(**{k: v for k, v in row.items() if k in columns}))

    def __iter__(self) -> Self:
        samples = []
        for idx, row in self.samples.iterrows():
            samples.append((_row := row.to_dict(), sample := self.input_fn(_row)))
            for key, value in sample.model_dump().items():
                column = f'!{key}'
                if column not in self.samples.columns:
                    self.samples[column] = None

                self.samples.at[idx, column] = value

        self._iter = iter(samples)
        self._count = 0
        return self

    def __next__(self) -> Tuple[Dict[str, Any], Schema]:
        if self._iter is None:
            raise StopIteration

        if self.max_samples and self._count >= self.max_samples:
            self._iter = None
            raise StopIteration

        self._count += 1
        return next(self._iter)

    def __len__(self) -> int:
        return len(self.samples)

    def save(self, path: Path) -> None:
        with open(path, 'wb') as f:
            dill.dump(self, f)

    @staticmethod
    def load(path: Path) -> "DatasetWrapper":
        with open(path, 'rb') as f:
            return dill.load(f)

    @staticmethod
    def from_hf(
            dataset: Dict[str, "Dataset"], # type: ignore
            input_fn: Optional[InputFn]=None,
            columns: Optional[List[str]]=None,
            max_samples: Optional[int]=None,
            expand: bool=False,
    ) -> Dict[str, "DatasetWrapper"]:
        from datasets import Dataset  # type: ignore[import-untyped]
        def expand_dataset(dataset: Dataset) -> Dataset:
            expanded_data = []
            for row in dataset:
                max_len = max(len(value) if isinstance(value, list) else 1 for value in row.values())
                for i in range(max_len):
                    new_row = {
                        key: value[i] if isinstance(value, list) and i < len(value) else value
                        for key, value in row.items()
                    }

                    for key, value in new_row.items():
                        if isinstance(value, list):
                            new_row[key] = str(value)
                    expanded_data.append(new_row)

            expanded_dict = {key: [row[key] for row in expanded_data] for key in expanded_data[0]}
            return Dataset.from_dict(expanded_dict)

        return {k: DatasetWrapper(expand_dataset(v) if expand else v, input_fn, columns, max_samples) for k, v in dataset.items()}

    @staticmethod
    def from_pandas(
            df: "pd.DataFrame", # type: ignore
            input_fn: Optional[InputFn]=None,
            columns: Optional[List[str]]=None,
            split_column: Optional[str]=None,
            max_samples: Optional[int]=None,
    ):
        from datasets import Dataset  # type: ignore[import-untyped]
        if split_column is not None:
            return DatasetWrapper.from_hf({
                str(split): Dataset.from_pandas(_df)
                    for split, _df in df.groupby(split_column)
            }, input_fn, columns, max_samples)

        return DatasetWrapper.from_hf({'all': Dataset.from_pandas(df)}, input_fn, columns, max_samples)



from verdict.prompt import Prompt


class GEvalEvaluationStepsCoTPrompt(Prompt):
    """
        {root.task}

        Evaluation Criteria:
        {criteria.name} ({criteria.scale}) - {criteria.description}

        Evaluation Steps:
    """



import string
from typing import List, Optional

from verdict.core.primitive import Unit
from verdict.scale import DiscreteScale
from verdict.schema import Field, Schema


class RankerUnit(Unit):
    _char: str = "Ranker"

    class InputSchema(Schema):
        # need a sentinel for empty, but using [''] for now makes type inference simple
        options: List[str] = Field(default=[''])

    class ResponseSchema(Schema):
        ranking: List[str]
    
    class OutputSchema(Schema):
        ranked_options: List[str]

    def __init__(self, k: int=3, options: Optional[DiscreteScale]=None, explanation: bool=False, original: bool=False, **kwargs):
        self.k = k
        self.explanation = explanation
        self.original = original

        options = options or DiscreteScale(list(string.ascii_uppercase[:k]))
        self.scale = options

        if explanation:
            self.ResponseSchema = self.ResponseSchema.prepend(explanation=str) # type: ignore
            self.OutputSchema = self.OutputSchema.prepend(explanation=str) # type: ignore

        super().__init__(**kwargs)
    
    def validate(self, input: InputSchema, response: ResponseSchema) -> bool:
        if self.original:
            assert len(input.options) != [''], "Pass the options in InputSchema.options to index original choices."
            assert len(input.options) == self.k, f"Number of input options must equal {self.k}"

    def process(self, input: InputSchema, response: ResponseSchema) -> OutputSchema:
        if not self.original:
            return response

        fields = {
            'ranked_options': [input.options[self.scale.values.index(self.scale.value_mapping_fn(rank))] for rank in response.ranking],
        }

        if self.explanation:
            fields['explanation'] = response.explanation

        return self.OutputSchema(**fields)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from verdict.model import ModelSelectionPolicy
from verdict.util.ratelimit import (ConcurrentRateLimiter, RateLimitPolicy,
                                    TimeWindowRateLimiter)


# Global state
@dataclass
class Config:
    rate_limiter_disabled: bool = False
state = Config()

# Defaults
## Rate limiting
PROVIDER_RATE_LIMITER: Dict[str, RateLimitPolicy] = {
    'deepinfra': RateLimitPolicy({
        ConcurrentRateLimiter(max_concurrent=200): 'requests'
    }),
    'together_ai': RateLimitPolicy.of(rpm=600, tpm=180_000),
    'openai': RateLimitPolicy({ # tier 1 for gpt-4o-mini
        TimeWindowRateLimiter(max_value=5, window_seconds=60): 'requests',
        TimeWindowRateLimiter(max_value=10_000, window_seconds=60*60*24): 'requests',
        TimeWindowRateLimiter(max_value=200_000, window_seconds=60): 'tokens'
    }),
    # TODO: add other providers
}

DEFAULT_RATE_LIMITER: RateLimitPolicy = PROVIDER_RATE_LIMITER['openai']

## Connection parameters
DEFAULT_PROVIDER_TIMEOUT: int = 120
DEFAULT_PROVIDER_STREAM_TIMEOUT: int = 120

## Inference parameters
DEFAULT_MODEL_SELECTION_POLICY: ModelSelectionPolicy = ModelSelectionPolicy.from_name('gpt-4o-mini', retries=3)
DEFAULT_INFERENCE_PARAMS: dict[str, Any] = {}

## Token extraction prompt
TOKEN_EXTRACTOR_SPECIFICATION_PROMPT: str = """
{content}
--------------------------------

Do not explain your answer. {scale_prompt}
"""

# System
## Logging
DEBUG: bool = bool(os.getenv('DEBUG', False))

LIGHTWEIGHT_EXECUTOR_WORKER_COUNT: int = 32

VERDICT_LOG_DIR: Path = Path.cwd() / '.verdict'
VERDICT_LOG_DIR.mkdir(parents=True, exist_ok=True)



from __future__ import annotations

import logging
import os
import signal
import sys
import threading
from functools import wraps
from typing import Any, Callable, Optional, Type


def is_signal_safe():
    """
    Strictly check if the current environment is safe for signal handling.

    Returns:
        bool: True only if ALL signal safety conditions are met
    """
    # Must be in the main thread
    if threading.current_thread() is not threading.main_thread():
        return False

    # Must be in the main interpreter
    if not hasattr(sys, 'argv'):
        return False

    # Additional interpreter state checks
    if not sys.modules.get('__main__'):
        return False

    # Check if the interpreter is shutting down
    if sys is None or threading is None:
        return False

    return True

def keyboard_interrupt_safe(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    @wraps(func)
    def wrapped(self: Any, *args, **kwargs) -> Any:
        original_handler = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, lambda signum, frame: self.executor.graceful_shutdown())

        try:
            return func(self, *args, **kwargs)
        except KeyboardInterrupt:
            self.executor.graceful_shutdown()
        finally:
            signal.signal(signal.SIGINT, original_handler)

    if is_signal_safe():
        return wrapped
    else:
        return func

class DisableLogger:
    def __init__(self, logger_name: str, all: bool=False) -> None:
        self.logger = logging.getLogger(logger_name)
        self.previous_level: Optional[int] = None
        self.all = all

    def __enter__(self) -> None:
        self.previous_level = self.logger.level
        new_level = logging.CRITICAL
        if self.all:
            new_level += 1
        self.logger.setLevel(new_level)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.logger.setLevel(self.previous_level) # type: ignore


def lightweight(cls: Type) -> Type:
    cls.lightweight = True
    return cls




import string
from typing import List, Optional

from verdict.core.primitive import Unit
from verdict.scale import DiscreteScale, LikertScale, Scale
from verdict.schema import Field, Schema


class JudgeUnit(Unit):
    """
    Direct score judge.
    """

    _char: str = "DirectScoreJudge"

    class ResponseSchema(Schema):
        score: Scale = LikertScale()

    def __init__(self, scale: Optional[Scale]=None, explanation: bool=False, **kwargs):
        if scale is not None:
            self.OutputSchema = self.ResponseSchema = Schema.infer(score=scale) # type: ignore

        self.scale = self.ResponseSchema.get_scale("score")

        if explanation:
            self.OutputSchema = self.ResponseSchema = self.ResponseSchema.prepend(explanation=str) # type: ignore

        super().__init__(**kwargs)

class BestOfKJudgeUnit(Unit):
    _char: str = "BestOfKJudge"

    class InputSchema(Schema):
        # need a sentinel for empty, but using [''] for now makes type inference simple
        options: List[str] = Field(default=[''])

    class ResponseSchema(Schema):
        choice: DiscreteScale = DiscreteScale(['A', 'B'])

    class OutputSchema(Schema):
        chosen: str
    
    def __init__(self, k: int=2, response_options: Optional[DiscreteScale]=None, explanation: bool=False, original: bool=False, **kwargs):
        self.k = k
        self.explanation = explanation
        self.original = original

        response_options = response_options or DiscreteScale(list(string.ascii_uppercase[:k]))
        self.ResponseSchema = Schema.infer(choice=response_options) # type: ignore

        self.scale = self.ResponseSchema.get_scale("choice")

        if explanation:
            self.ResponseSchema = self.ResponseSchema.prepend(explanation=str) # type: ignore
            self.OutputSchema = self.OutputSchema.prepend(explanation=str) # type: ignore

        super().__init__(**kwargs)

    def validate(self, input: InputSchema, response: ResponseSchema) -> bool:
        if self.original:
            assert len(input.options) != [''], "Pass the options in InputSchema.options to index original choices."
            assert len(input.options) == self.k, f"Number of input options must equal {self.k}"

    def process(self, input: InputSchema, response: ResponseSchema) -> OutputSchema:
        if not self.original:
            return response

        fields = {'chosen': input.options[self.scale.values.index(self.scale.value_mapping_fn(response.choice))]}
        if self.explanation:
            fields['explanation'] = response.explanation

        return self.OutputSchema(**fields)

class PairwiseJudgeUnit(BestOfKJudgeUnit):
    """
    Judge unit for pairwise comparison (e.g., 'Response A' vs 'Response B').
    Options should be LLM responses.
    """

    _char: str = "PairwiseJudge"

    def __init__(self, response_options: Optional[DiscreteScale]=None, explanation: bool=False, original: bool=False, **kwargs):
        super().__init__(k=2, response_options=response_options, explanation=explanation, original=original, **kwargs)


class CategoricalJudgeUnit(BestOfKJudgeUnit):
    """
    Judge unit for categorical decisions (e.g., 'Yes' or 'No', 'Harmful' or 'Not Harmful', 'Hallucination' or 'No Hallucination').
    Options should represent discrete categorical choices.
    """

    _char: str = "CategoricalJudge"

    def __init__(self, categories: Optional[DiscreteScale]=None, explanation: bool=False, original: bool=False, **kwargs):
        super().__init__(k=2, response_options=categories, explanation=explanation, original=original, **kwargs)



import ast
import inspect
import random
import re
import string
from abc import ABC, abstractmethod
from typing import (Any, Dict, List, NamedTuple, Optional, Set, Tuple, Type,
                    Union)

from loguru._logger import Logger
from typing_extensions import Self

from verdict.schema import Schema
from verdict.util.exceptions import PromptError

SECTION_REGEX = re.compile(r'@(\w+)(.*?)(?=@\w+|$)', re.DOTALL)

class PromptRegistry(type):
    _registry: Dict[str, Type["Prompt"]] = {}

    @staticmethod
    def extract_sections(template: str) -> Tuple[Optional[str], Optional[str], bool]:
        matches = SECTION_REGEX.findall(template)

        sections = {f"@{section}": content for section, content in matches}

        # Assign leading text to '@user' only if it exists and '@user' is not explicitly defined
        leading_text = template.split('@', 1)[0]
        if leading_text and '@user' not in sections:
            sections['@user'] = leading_text

        return sections.get('@system'), sections.get('@user'), '@no_format' in sections

    @staticmethod
    def strip_prompt_template(prompt: str) -> str:
        lines = prompt.splitlines()

        # Remove leading empty lines
        while lines and not lines[0].strip():
            lines.pop(0)

        # Remove trailing empty lines
        while lines and not lines[-1].strip():
            lines.pop()

        # Unindent the remaining lines and reassemble the docstring
        return "\n".join(line.lstrip() for line in lines)

    def __new__(cls, name, bases, dct) -> Type["Prompt"]:
        klass = super().__new__(cls, name, bases, dct)

        # only add named classes to the registry
        if name and not name.startswith('_') and name != 'Prompt':
            cls._registry[name] = klass

        if not klass.__doc__:
            raise PromptError(f"Prompt {klass.__name__} must have __doc__.")

        if '' in (klass.extract_keys(klass.__doc__)):
            raise PromptError(f"All keys must be named in Prompt {klass.__name__}.")

        klass.system_prompt_template, klass.user_prompt_template, no_format = klass.extract_sections(klass.__doc__)
        assert klass.user_prompt_template is not None, "User prompt template must be specified."

        if not no_format:
            klass.user_prompt_template = klass.strip_prompt_template(klass.user_prompt_template)
            if klass.system_prompt_template:
                klass.system_prompt_template = klass.strip_prompt_template(klass.system_prompt_template)

        return klass

    def __getattr__(self, name: str) -> Type["Prompt"]:
        if name in self._registry:
            return self._registry[name]
        else:
            raise PromptError(f"Prompt {name} not found in PromptRegistry._registry.")

    @staticmethod
    def compatible_prompts(unit: Type["Promptable"]) -> List[Type["Prompt"]]:
        return [prompt for prompt in PromptRegistry._registry.values() if prompt().supports(unit)]

class PromptMessage(NamedTuple):
    system: Optional[str]
    user: str

    def to_messages(self, add_nonce: bool=False) -> List[Dict[str, str]]:
        nonce = ''.join(random.choices(string.ascii_letters, k=10)) + '\n' if add_nonce else ''
        messages = [{"role": "user", "content": nonce + self.user}]
        if self.system:
            messages.insert(0, {"role": "system", "content": nonce + self.system})
        return messages

RESERVED_KEYS = set(["input", "unit", "previous", "source", "prompt"])
class Prompt(metaclass=PromptRegistry):
    """
    A prompt template that can be used to generate a prompt for an LLM.
    Supports formatting with the keys of a Schema / pydantic.BaseModel.

    {key} will be replaced with the value of the key in the input at runtime.
    Indentation is preserved in the output.
    """

    system_prompt_template: Optional[str]
    user_prompt_template: str

    caller_locals: Dict[str, Any] = {}

    @staticmethod
    def from_template(template: str) -> "Prompt":
        return type(
            "",
            (Prompt,),
            {"__doc__": template}
        )()

    @staticmethod
    def extract_keys(template: str, exclude_reserved: bool=True) -> Set[str]:
        def extract_variables(template: str) -> Set[str]:
            variables = set()

            placeholder_pattern = r"(?<!\{)\{([^{}]+?)\}(?!\})"
            placeholders = re.findall(placeholder_pattern, template)

            for placeholder in placeholders:
                try:
                    tree = ast.parse(placeholder, mode='eval')

                    class VariableVisitor(ast.NodeVisitor):
                        def visit_Name(self, node) -> None:
                            variables.add(node.id)
                        def visit_Attribute(self, node) -> None:
                            while isinstance(node, ast.Attribute):
                                node = node.value
                            if isinstance(node, ast.Name):
                                variables.add(node.id)

                    VariableVisitor().visit(tree)
                except SyntaxError:
                    variables.add(placeholder.split(".")[0])

            return variables

        if not exclude_reserved:
            return extract_variables(template)
        return extract_variables(template) - RESERVED_KEYS
    
    def get_all_keys(self) -> Set[str]:
        keys = set()

        if self.system_prompt_template:
            keys |= Prompt.extract_keys(self.system_prompt_template, exclude_reserved=False)

        keys |= Prompt.extract_keys(self.user_prompt_template, exclude_reserved=False)

        return keys

    @property
    def keys(self) -> Set[str]:
        return self.get_all_keys() - RESERVED_KEYS - set(self.caller_locals.keys())

    def format(
            self,
            input_schema: Schema,
            unit: "Promptable",
            previous: Any,
            source: Schema,
            logger: Optional[Logger]=None
        ) -> PromptMessage:
        for illegal_symbol in ['{{', '}}']:
            for prompt in [self.system_prompt_template, self.user_prompt_template]:
                if prompt is not None and illegal_symbol in prompt:
                    if logger:
                        logger.warning(f"Prompt contains '{illegal_symbol}'. Variable likely not getting evaluated.")

        format_kwargs = {"input": input_schema} \
                      | {"unit": unit} \
                      | {"previous": previous} \
                      | {"source": source}

        for key, value in self.caller_locals.items():
            if key not in format_kwargs and not key.startswith("__"):
                format_kwargs[key] = value

        return PromptMessage(
            system=self.auto_format(self.system_prompt_template, format_kwargs) if self.system_prompt_template else None,
            user=self.auto_format(self.user_prompt_template, format_kwargs)
        )

    @staticmethod
    def auto_format(template: str, context: Dict[str, Any]) -> str:
        single_placeholder_pattern = r"(?<!\{)\{([^{}]+?)\}(?!\})"
        matches = re.findall(single_placeholder_pattern, template)

        for match in matches:
            try:
                value = eval(match, {}, context)
                template = template.replace(f"{{{match}}}", str(value))
            except Exception as e:
                raise PromptError(f"""Failed to evaluate Prompt placeholder '{match}' in the following context.

Context: {context}""") from e

        return template


class Promptable(ABC):
    _prompt: Prompt

    def prompt(self, prompt: Union[str, Prompt]) -> Self:
        if isinstance(prompt, str):
            self._prompt = Prompt.from_template(prompt)
        else:
            self._prompt = prompt

        frame = inspect.currentframe().f_back
        self._prompt.caller_locals = frame.f_locals

        return self

    @abstractmethod
    def populate_prompt_message(self, input: Schema, logger: Logger) -> PromptMessage:
        pass



import statistics
from operator import attrgetter
from typing import Any, Callable, List, Union

from verdict.core.primitive import Unit
from verdict.schema import Schema
from verdict.util.exceptions import VerdictExecutionTimeError
from verdict.util.misc import lightweight


@lightweight
class MapUnit(Unit):
    _char: str = "Map"

    accumulate: bool = True
    map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]]

    def __init__(
        self,
        map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.data.map_func = lambda x: map_func(x)

    class InputSchema(Schema):
        values: Union[Any, List[Any]]

    class ResponseSchema(Schema):
        values: Union[Any, List[Any]]

    def execute(
        self, input: InputSchema
    ) -> ResponseSchema:
        try:
            values = input.values
            if len(values) == 1:
                values = values[0]

            output = self.data.map_func(values)
            if isinstance(output, Schema):
                self.OutputSchema = self.ResponseSchema = output.__class__ # type: ignore
                return output # type: ignore

            return MapUnit.ResponseSchema(values=output)
        except Exception as e:
            raise VerdictExecutionTimeError(
                f"Failed to execute MapUnit: {e} on {input}"
            ) from e

class FieldMapUnit(MapUnit):
    fields: List[str]

    def __init__(
        self,
        map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]],
        fields: Union[str, List[str]],
        **kwargs
    ):
        super().__init__(map_func, **kwargs)
        self.fields = fields if isinstance(fields, list) else [fields]

    def execute(self, input: MapUnit.InputSchema) -> MapUnit.ResponseSchema:
        values = input.values

        if len(self.fields) == 0 and len(values) > 0:
            self.fields = list(values[0].model_fields.keys())

        assert all(field in values[0].model_fields for field in self.fields), f"Fields {self.fields} not a subset of input {input.values}"
        return Schema.of(**{ # type: ignore
            field: self.data.map_func(
                list(map(attrgetter(field), values))
            ) for field in self.fields
        })

    @staticmethod
    def from_fn(fn: Callable[[List[Any]], Any], name: str) -> Callable[[Union[str, List[str]]], "FieldMapUnit"]:
        return lambda fields=[]: FieldMapUnit(fn, fields, name=name)


class MeanPoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(statistics.mean, fields, name="MeanPool")


class MaxPoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(statistics.mode, fields, name="MaxPool")


class MeanVariancePoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(
            lambda x: Schema.inline(mean=float, variance=float)(  # type: ignore
                mean=statistics.mean(x),
                variance=statistics.variance(x) if len(x) > 1 else 0.0,
            ),
            fields,
            name="MeanVariancePool",
        )


__all__ = [
    'MapUnit',

    'MeanPoolUnit',
    'MeanVariancePoolUnit',

    'MaxPoolUnit',
]
